import json
from operator import attrgetter
import pickle

from pybrary.model_base import Attribute, Model


def select(cls, **where):
    getters = [
        (attrgetter(k.replace('_', '.')), v)
        for k, v in where.items()
    ]
    found = list()
    for sub in cls.class_tree():
        for obj in sub.instances:
            for get, val in getters:
                try:
                    if get(obj) != val: break
                except AttributeError: break
            else:
                found.append(obj)
    return found


def backup(cls, path=None):
    path = path or '/tmp/model.pkl'
    models = {
        sub.__name__ : tuple(sub.instances)
        for sub in cls.class_tree()
    }
    with open(path, 'wb') as out:
        pickle.dump(models, out)


def restore(cls, path=None):
    path = path or '/tmp/model.pkl'
    with open(path, 'rb') as inp:
        models = pickle.load(inp)   # noqa: S301
    return models


def export(cls, path=None):
    path = f'{path or cls.__name__}.json'
    models = {
        sub.__name__ : [i.dump() for i in sub.instances]
        for sub in cls.class_tree()
        if sub.instances
    }
    with open(path, 'w') as out:
        json.dump(models, out, indent=4)
    return models


class ModelPlus(Model):
    def __init__(self, *args, **kw):
        super().__init__(*args, **kw)
        self.validate_attrs()

    def validate_attrs(self):
        for _name, attr in self.attrs.items():
            if isinstance(attr, Model):
                attr.validate()

    @classmethod
    def select(cls, **where):
        return select(cls, **where)

    @classmethod
    def backup(cls, path=None):
        backup(cls, path)

    @classmethod
    def restore(cls, path=None):
        return restore(cls, path)

    @classmethod
    def export(cls, path=None):
        return export(cls, path)


class Link(Attribute):
    def __init__(self, relation_name=None):
        self.relation_name = relation_name or object
        self._relation_ = None

    @property
    def relation(self):
        if not self._relation_:
            name = self.relation_name
            # if isinstance(name, type) and issubclass(name, Model):
            if isinstance(name, type):
                self._relation_ = name
            elif isinstance(name, str):
                for cls in Model.class_tree():
                    if cls.__name__ == name:
                        self._relation_ = cls
                        break
                else:
                    raise ValueError(f'Relation not found: {name}')
            else:
                raise ValueError(f'Invalid relation type : {type(name).__name__}')
        return self._relation_


class One(Link):
    def __set__(self, obj, val):
        if isinstance(val, self.relation):
            relation = val
        elif isinstance(val, dict):
            relation = self.relation(**val)
        elif isinstance(val, str):
            relation = self.relation(val)
        else:
            self.invalid(val, f'should be a {self.relation.__name__}')
        self.set(obj, relation)
        if isinstance(relation, Model):
            setattr(relation, type(obj).__name__.lower(), obj)


class Many(Link):
    def __init__(self, *a, **k):
        super().__init__(*a, **k)
        self.items = list()

    def __get__(self, obj, cls):
        self._obj_ = obj
        return self

    def __set__(self, obj, val):
        raise ValueError(f'{self.name} is a list and cannot be set.')

    def __iter__(self):
        yield from self.items

    def add(self, relation):
        if not isinstance(relation, self.relation):
            self.invalid(relation, f'should be a {self.relation.__name__}')
        self.items.append(relation)
        if isinstance(relation, Model):
            attr_name = self._obj_.attr_name if hasattr(self._obj_, 'attr_name') else type(self._obj_).__name__.lower()
            setattr(relation, attr_name, self._obj_)

    def __repr__(self):
        if self.items:
            return ', '.join(str(i) for i in self.items if i)
        return ''

    def dump(self, val):
        return self.items




import json
from weakref import WeakSet


class Attribute:
    default = None
    mandatory = False

    def __init__(self, default=None, mandatory=False):
        # attr values or desc values
        self.default = default or self.default
        self.mandatory = mandatory or self.mandatory

    def __set_name__(self, cls, name):
        self.name = f'{cls.__name__}.{name}'
        self.attr = f'_{name}_'

    def __get__(self, obj, cls):
        return self.get(obj)

    def __set__(self, obj, val):
        try:
            validated = self.validate(val)
        except ValueError:
            self.invalid(val)
        if validated:
            self.set(obj, validated)
        else:
            self.invalid(val)

    def set(self, obj, val):
        setattr(obj, self.attr, val)

    def get(self, obj):
        return getattr(obj, self.attr, self.default)

    def validate(self, val):
        return val

    def invalid(self, val, msg=None):
        msg = f' ; {msg}' if msg else ''
        raise ValueError(f'Invalid {self.name} {type(val).__name__}({val}){msg}.')

    def dump(self, val):
        return val


class Property(Attribute):
    def __set__(self, obj, val):
        raise ValueError(f'{self.name} is read only.')


def clear(cls):
    cls.instances = WeakSet()
    for sub in cls.class_tree():
        sub.instances = WeakSet()


class Meta(type):
    def __init__(cls, name, bases, ns):
        super().__init__(name, bases, ns)
        cls.instances = WeakSet()


class Model(metaclass=Meta):
    _sentinel_ = None

    def __new__(cls, *a, **k):
        instance = super().__new__(cls)
        cls.instances.add(instance)
        return instance

    def __init__(self, *args, **kw):
        self.set_attrs(*args, **kw)
        self.validate()

    @property
    def lineage(self):
        return list(reversed(self.__class__.mro()[:-2]))

    def get_attrs(self):
        '''Get Model's Attributes.
        '''
        for cls in list(reversed(self.__class__.mro()))[2:]:
            for attr, desc in cls.__dict__.items():
                if attr.startswith('_'): continue
                if isinstance(desc, Attribute) and not isinstance(desc, Property):
                    yield attr, desc

    def set_attrs(self, *args, **kw):
        '''Set Attributes values from args.
        '''
        idx = 0
        for attr, desc in self.get_attrs():
            val = kw.pop(attr, None)
            if not val:
                try:
                    val = args[idx]
                except IndexError:
                    if desc.mandatory:
                        raise ValueError(f'Attribute {desc.name} is mandatory.') from None
                    # optional arg, will use default value.
                    continue
                idx += 1
            setattr(self, attr, val)

        if len(args) > idx:
            args = [str(a) for a in args[idx:]]
            raise ValueError(f'Invalid positional arguments for {self.__class__.__name__} : {", ".join(args)}')

        if kw:
            kw = [f'{k}={v}' for k, v in kw.items()]
            raise ValueError(f'Invalid keyword arguments for {self.__class__.__name__} : {", ".join(kw)}')

    @property
    def attrs(self):
        attrs = {
            attr : getattr(self, attr)
            for attr, _ in self.get_attrs()
        }
        return attrs

    @property
    def dict(self):
        items = {
            name : getattr(attr, 'attrs', attr)
            for name, attr in self.attrs.items()
        }
        return items

    @classmethod
    def class_tree(cls):
        yield cls
        for sub in cls.__subclasses__():
            yield from sub.class_tree()

    @classmethod
    def clear(cls):
        clear(cls)

    def dump(self):
        if self._sentinel_ is self:
            return '_self_'
        self._sentinel_ = self
        MODEL = self.__class__.mro()[1]   # noqa: N806
        data = dict()
        for name, desc in self.get_attrs():
            if attr := getattr(self, name):
                if isinstance(attr, MODEL):
                    val = attr.dump()
                    if val == '_self_':
                        continue
                else:
                    val = desc.dump(attr)
                if val:
                    if isinstance(val, Model):
                        val = val.dump()
                    data[name] = val
        self._sentinel_ = None
        return data

    def json(self, path='model'):
        with open(f'{path}.json', 'w') as out:
            data = self.dump()
            json.dump(data, out, indent=4)
        return data

    def __repr__(self):
        if self._sentinel_ is self:
            return '_self_'
        self._sentinel_ = self
        attrs = [
            (k, repr(v))
            for k, v in self.attrs.items()
            if v and repr(v) != '_self_'
        ]
        attrs = [
            (k, r)
            for k, r in attrs
            if r and r != '_self_'
        ]
        attrs = ', '.join(f'{key}={val}' for key, val in attrs)
        self._sentinel_ = None
        return f'{self.__class__.__name__}({attrs})'

    def validate(self):
        '''Validate Model after initialization.

        To be overwritten if needed.
        '''

